// {{{  Copyright (C) 2009 Nuno David Lopes.-----------------------------------
// e-mail: ndl@ptmat.fc.ul.pt
// http://ptmat.fc.ul.pt/~ndl/
// ---------------------------------------------------------------------------
// Licensed under the GNU LGPL Version 3.0 or later.
// ---------------------------------------------------------------------------
//  main.cpp
//  This program provides a demo of the DOLFWAVE library. 
//  small scale toy model of harbour note (g=9.81m/s^2) (domain scale=45m) 
// --------------------------------------------------------------------------
// Using Viper  pr post-processor and Xd3d with the files output.
// }}}------------------------------------------------------------------------
#include <dolfwave.h>
using namespace dolfin;
using namespace dolfin::dolfwave;
// {{{ User defined Functions---------------------------------------------------
class DirichletBoundary : public SubDomain
  {
    bool inside(const double* x, bool on_boundary) const
    {
      return 
        (x[0] < -14.99) &&
        on_boundary;
    }
  };
// Dirichlet Boundary conditions
class DBC : public Function
{
public:
  DBC(const double& t) : t(t)
  {}
private:
  void eval(double* values,const double* x) const
  {
    double time=t;
    double a=0.01;
    double omega=3.1142485275;
    double b=0.044861357;
    double k=1.6845;
    values[0]= a*sin(-omega*time);
    values[1]=-(b/k)*cos(-omega*time)+b/k;
  };
  const double& t;
};
//Depth function
class Depth : public Function
{
  void eval(double* values,const double* x) const
  {
    double retrn=0.0;

    if (x[0]<10.0)
      retrn=0.4572;
    else if 
      ((x[0]>= 10.0)&&
       (x[0]<= 17.0))
      retrn=0.4572+(10.-x[0])/25.0;
    else 
      retrn=0.1772;

    values[0]=retrn; 
  }
};
//Sponge Layer definition
class SpongeLayer : public Function
{
  void eval(double* values,const double* x) const
  {
    double retrn=0.0;
    double n1=1.5;
    double n2=0.01;
    double xs=17.;
    double x2=22.;
    if (x[0]>xs) 
      retrn=n1*(std::pow(exp((x[0]-xs)/(x2-xs)),n2)-1.)/(exp(1.)-1.); 
    
    values[0]=retrn;
  }
};
// }}}End User defined Functions ----------------------------------------------
int main()
{
  // Dolfwave start
  Dolfwave dw(30000  /*MaxSteps*/,
              0.0004 /*TimeStep*/,
              20 /*WriteGap*/,
              "../../meshes/harbour/harbour.xml.gz" /*MeshName*/,
              1.0e-8 /*PcEpsilon (<=0, one step >0 multi-step corrector)*/,
              "Zhao" /*FormName*/,
              "uBLAS"/* AlgBackEND*/ ,
              "viper" /*Plotter*/,
              "output/eta.raw" /*EtaFile*/,
              "output/phi.raw"/*PhiFile*/,
              "output/phyphi.raw" /*PhyPhiFile*/,
              "output/horvel.raw" /*HorVelFile*/); 
  //  Mesh and boundary definition
  Mesh mesh(dw.MeshName);
  mesh.order();
  Constant g(9.81); //gravity acceleration
  Depth depth;//Depth
  SpongeLayer sponge;//Sponge Layer
  Constant zero(0.0);//Source Function
  dw.FunctionSpaceInit(mesh);
    // Boundary conditions definition  
  DirichletBoundary dirichletboundary;
  DBC bc(dw.Time);
  dw.DepthPlot(mesh,depth);//Save depth
  dw.SpongePlot(mesh,sponge);//Save sponge
  dw.BilinearFormInit(depth);//Bilinear forms init
  dw.MatricesAssemble( );  //Matrices assemble and Dir. Bc apply 
  dw.FunctionsInit( );//Unknown Functions init 
  dw.LinearFormsInit(depth,g,
                     sponge /*sp_eta*/,zero /*sp_lap_eta*/,
                     zero   /*sp_phi*/,zero /*sp_lap_phi */,
                     zero /*source function*/);
  dw.VectorInit( );//Aux Vector initialization
  dw.DolfwaveDBC(bc,dirichletboundary); //Setting Dirichlet Boundary Conditions    
  dw.LinearSolver( );//Linear Solver //limited yet
  dw.RKSolve( );
  dw.PCInit();  
  for(dolfin::uint i=4; i<dw.MaxSteps+1;i++)
    {
      dw.PCSolve( );    
      //  Plot information------------------------
      if(!(i%dw.WriteGap))
        dw.Plot(true, false,false /*SaveEta*/,false /*SavePhi*/);
      message("Step=%d \t Simulated time= %.4f",i,dw.Time);
    }
  return (EXIT_SUCCESS);
}
