// {{{  Copyright (C) 2009 Nuno David Lopes.-----------------------------------
//
// e-mail: ndl@ptmat.fc.ul.pt
// http://ptmat.fc.ul.pt/~ndl/
//
// ---------------------------------------------------------------------------
// Licensed under the GNU LGPL Version 3.0 or later.
//
// ---------------------------------------------------------------------------
//  main.cpp
//
//  This program provides a demo of the DOLFWAVE library. 
//
// --------------------------------------------------------------------------
// Using Viper  pr post-processor and Xd3d with the files output.
// }}}------------------------------------------------------------------------

// {{{ Includes --------------------------------------------------------------
#include <fstream> //for ofstream File saving
#include <string> //for string 
#include "../../src/dolfwave.h"
// }}} -----------------------------------------------------------------------

// {{{  Parameteres and global variables---------------------------
const dolfin::uint MaxSteps=12000;
const double TimeStep=10.;
const dolfin::uint WriteGap=100; 
const double PcEpsilon=1.0e-8;
const std::string FormName="Zhao";
const std::string MeshName="../../meshes/harbour/real_harbour.xml.gz";
// }}} ----------------------------------------------------
using namespace dolfin;
using namespace dolfin::dolfwave;

// {{{ User defined Functions--------------------------------------------------------------
class DirichletBoundary : public SubDomain
  {
    bool inside(const double* x, bool on_boundary) const
    {
      return 
        (x[0] < -449.) &&
        on_boundary;
    }
  };

// Dirichlet Boundary conditions
class DBC : public Function
{
public:
  DBC(const double& t) : t(t)
  {}
private:
  void eval(double* values,const double* x) const
  {
    double time=t;
    double a=0.3; /* amplitude 0.3m-> 0.6 wave height */
    double omega=1.; /* period 2*Pi \approx 7s */
    
    double k=0.1088; /* wave number -> wave lenght related */   
    double c=0.4062; /*wave velocity magnitude */
  
    /* note: b=c/k potencial magnitude*/
    /*b\approx a*omega(1+(2/5)*(kH)^2/(k^2 H))-> from dispersion relation */

    values[0]= a*sin(-omega*time);
    values[1]=-(c/k)*cos(-omega*time)+c/k;
  };
  const double& t;
};

//Depth function
class Depth : public Function
{
  void eval(double* values,const double* x) const
  {
    double retrn=0.0;

    if (x[0]<300.0)
      retrn=13.716;
    else if 
      ((x[0]>= 300.0)&&
       (x[0]<= 510.0))
      retrn=13.716+(300.-x[0])/25.0;
    else 
      retrn=5.316;

    values[0]=retrn; 
  }
};

//Sponge Layer definition
class SpongeLayer : public Function
{
  void eval(double* values,const double* x) const
  {
    double n1=1.5;
    double n2=0.05;
    double n3=0.02;
    double xs=510.;
    double x2=660.;
    values[0]=0.0;
   
    double centre1x=-120.0;
    double centre1y=300.0;

    double centre2x=-120.0;
    double centre2y=-120.0;
    
    if (x[0]>xs) 
      values[0]+=n1*(std::pow(exp((x[0]-xs)/(x2-xs)),n2)-1.)/(exp(1.)-1.); 
    
    if (std::sqrt(sqr(x[0]-centre1x)+sqr(x[1]-centre1y))<150.0)
      values[0]+=n1*(std::pow(exp(
                                  (150.0-std::sqrt(sqr(x[0]-centre1x)+sqr(x[1]-centre1y)))/150.0),n3)-1.)/(exp(1.)-1.);
   
    if (std::sqrt(sqr(x[0]-centre2x)+sqr(x[1]-centre2y))<120.0)
      values[0]+=n1*(std::pow(exp(
                                  (120.0-std::sqrt(sqr(x[0]-centre2x)+sqr(x[1]-centre2y)))/120.0),n3)-1.)/(exp(1.)-1.);
  }
};

// }}}End User defined Functions --------------------------------------------------

int main()
{
  // {{{  Clear and maintimer.start()
  Dolfwave dw(MaxSteps,TimeStep,WriteGap,MeshName,PcEpsilon,FormName); 
  Timer maintimer("Process execution");
  maintimer.start();    
  // }}}

  // {{{User dependent information
  //  Mesh and boundary definition
  Mesh mesh(dw.MeshName);
  mesh.order();

  // Functions
  /***********************************************/
  /* Rescale the gravity accelaration ************/ 
  Constant g(9.81); //gravity acceleration $m/s^2$
  /***********************************************/

  Depth depth;//Depth
  SpongeLayer sponge;//Sponge Layer
  Constant src(0.0);//Source Function
  FunctionSpace *V;
  SubSpace *V0,*V1;
  FunctionSpaceInit(dw.FormName,mesh,V,V0,V1);
  // 
 
  // Boundary conditions definition
  double t=0.0; //time variable
  
  DirichletBoundary dirichletboundary;
  DBC bc(t);
  DirichletBC dbc(*V,bc,dirichletboundary);  
  // }}}
  
  // {{{ DEPTH and SPONGE PLOT ------------------------------------------
  dw.DepthPlot(mesh,depth);
  dw.SpongePlot(mesh,sponge); 
  // }}} DEPTH PLOT ------------------------------------------
  
  // {{{ Functions, Vectors and Matrices ----------------------
  Matrix M;
  Form *a,**L;
  BilinearFormInit(dw.FormName,*V,a,depth);
  assemble(M,*a);  
  
  //Solutions Function and subfunctions (potential,elevation)
  Function **eta_phi,**eta,**phi;
  FunctionsInit(eta_phi,eta,phi,*V,*V0,*V1); 
  
  LinearFormsInit(dw.FormName,*V,L,eta,phi,depth,g,sponge,src);
 
  Vector *pcrk_b;
  VectorInit(pcrk_b,eta_phi);
  // }}}---------------------------------------------------------
 
  // {{{ Runge-Kutta for the initialization of the 3 initial steps
  RungeKutta rk(RungeKutta::euler);
  rk.disp();
  LUSolver lu;
  Matrix Mf(M);
  dbc.apply(Mf,pcrk_b[5]);
  lu.factorize(Mf);

  rk.solve(pcrk_b,L,M,eta_phi,eta,phi,dw.TimeStep,dbc,lu,Mf,t);
  // }}}  
  
  PredCorr pc(*V0,eta_phi);
  
  Timer pc_looptimer("Predictor-Corrector solver");
   
  for(dolfin::uint i=4; i<dw.MaxSteps+1;i++)
    {
      //One step corrector
      pc.solve(pcrk_b,L,M,eta_phi,eta,phi,
               dw.TimeStep,dbc,lu,Mf,t);    
     
      t+=dw.TimeStep;
      
      //  Plot information------------------------
      dw.Plot(i,dw.WriteGap,eta_phi,false,dw.SaveEta,dw.SaveEta);
      dw.PhyPhiPlot(i,dw.WriteGap,mesh,depth,phi,false,dw.SavePhyPhi);
      dw.HorVelPlot(i,dw.WriteGap,mesh,phi,false,dw.SaveHorVel);
      message("Step=%d \t Simulated time= %.4f",i,t);
    }
  maintimer.stop();
  summary();
  return (EXIT_SUCCESS);
}
