// {{{ Copyright (C) 2009 Nuno David Lopes.-----------------------------------
//
// e-mail: ndl@ptmat.fc.ul.pt
// http://ptmat.fc.ul.pt/~ndl/
//
// ---------------------------------------------------------------------------
// Licensed under the GNU LGPL Version 3.0 or later.
//
// ---------------------------------------------------------------------------
//  dolfwave.h
//
// First added:  2009-04-14
// Compiled with DOLFIN:0.9.2
// }}}------------------------------------------------------------------------


#ifndef __DOLWAVECLASS_H
#define __DOLFWAVECLASS_H

#include <dolfin.h>
#include <fstream> //for ofstream File saving
#include <string> //for string 

#include "forms/Zhao.h"
#include "forms/L2Norm.h"
#include "forms/L2Proj.h"
#include "forms/L2ProjP2.h"
#include "forms/ZhaoPotentialTrans.h"
#include "forms/LPT2z0.h"
#include "forms/LPTcdgI.h"


#include"RungeKutta.h"
#include"PredictorCorrector.h"

namespace dolfin
{
  namespace dolfwave
  {


    class Dolfwave
    {
    public:
      
      // {{{ Constructors-------------------------------------------------------
      Dolfwave( );//Useless
      /* Not implemented yet*/      
      Dolfwave(std::string FILENAME="dolfwave.txt"); //Read parameters from file
      /*Explicit constructor */
      Dolfwave(dolfin::uint MAXSTEPS ,
               double TIMESTEP ,
               dolfin::uint WRITEGAP,
               std::string MESHNAME,
               double PCEPSILON=1.0e-8,
               std::string FORMNAME="LPT2z0",
               std::string ALGBACKEND="uBLAS",
               std::string PLOTTER="viper",
               std::string ETAFILE="output/eta.raw",
               std::string PHIFILE="output/phi.raw",
               std::string PHYPHIFILE="output/phyphi.raw",
               std::string HORVELFILE="output/horvel.raw");
      // }}}--------------------------------------------------------------------
      
      // {{{ Destructor and AbortDolfwave------------------------------------------
      inline ~Dolfwave(){TimerStop();exit (EXIT_SUCCESS);};
      inline void AbortDolfwave(){TimerStop();fputs("FAILUREDOLFWAVE",stderr);abort();};                          
      // }}} -------------------------------------------------------------------
      
      // {{{ Options to be set by constructor----------------------------------- 
      dolfin::uint MaxSteps;
      double TimeStep;
      dolfin::uint WriteGap;
      std::string MeshName; 
      double PcEpsilon;
      std::string FormName;
      std::string AlgBackEnd;
      std::string Plotter;
      std::string EtaFile;
      std::string PhiFile; 
      std::string PhyPhiFile;
      std::string HorVelFile;
      // }}}-------------------------------------------------------------------- 

      // {{{ Time related members----------------------------------------------- 
      double Time; 
      void TimeUpdate( ); 
      void TimeAdd(double );// Change Time
      //}}}---------------------------------------------------------------------
  
      // {{{Plot and info functions----------------------------------------------
      void disp(); //Display options
      void Plot( const bool show_eta=true,const bool show_phi=true,
                 const bool surface=true,const bool potential=false) const; //preview and/or save eta and phi
      void ViperPlot(const dolfin::Function &) const;
      void Xd3dXyzPlot(const dolfin::Function &) const;
      void DepthPlot(const Mesh & ,Function &,const bool show=false) const; 
      void SpongePlot(const Mesh & ,Function &,const bool show=false) const; 
      void HorVelPlot(const Mesh & ,Function**& ,const bool show=false, const bool save=true) const;      
      void PhyPhiPlot(const Mesh & ,Function &, Function**&  ,Function &,
                      const bool show=false, const bool save=true) const;    
      void PointPlot(const double ,const double, Function**&,const std::string ) const;
      // }}}---------------------------------------------------------------------
      
      // {{{Functions,Spaces,Subspaces,forms and Vectors------------------------
      FunctionSpace *V;//Spaces and Subspaces
      SubSpace *V0,*V1;
      Form *a;// Bilinear and Linear Forms
      Form **L;
      Function **eta_phi,**eta,**phi;// Unknown Functions
      Vector *pcrk_b;//Auxiliar Vector
      // }}}---------------------------------------------------------------------
      
      // {{{ Dolfwave Problem set and solve Functions ---------------------------
      void FunctionSpaceInit(dolfin::Mesh &);
      void BilinearFormInit(dolfin::Function &);
      void TestBilinearFormInit(dolfin::Function &,dolfin::Function &,
                            dolfin::Function &,dolfin::Function & );
      void LinearFormsInit(dolfin::Function &,dolfin::Function &,
                           dolfin::Function &,dolfin::Function &);
      void LinearFormsInit(dolfin::Function &,dolfin::Function &,
                           dolfin::Function &,dolfin::Function &, 
                           dolfin::Function &,dolfin::Function &,
                           dolfin::Function &);
      
      void TestLinearFormsInit(dolfin::Function &, dolfin::Function &,
                               dolfin::Function &, dolfin::Function &,
                               dolfin::Function &);
      
      void FunctionsInit( );//Unkown functions Init
      void VectorInit( );// Auxiliar Vectors Initialization      
      void InitialCondition(Function&,Function&);
      void InitialCondition(std::string ,std::string );
      void MatricesAssemble(  );
      void LinearSolver( );
      LUSolver lu;// TODO Have decide about LinearSolvers 
      /*Runge-Kutta do not need an init since 
        it is only called once*/
      void RKSolve( );  
      void PCInit( );
      void PCSolve( );  
      
      //DirichletBoundary Conditions
      void DolfwaveDBC(dolfin::Function &, dolfin::SubDomain & );
      // }}}--------------------------------------------------------------------

      //Problem Matrices
      Matrix *M;  // left-hand side
      Matrix *Mf; // right-hand side
      
     
    private:
      //Predictor-Corrector pointer
      PredCorr *pc;

      // {{{ Timer functions----------------------------------------------------
      dolfin::Timer *maintimer;
      void TimerStop( ); //Used by destructor
      void TimerStart(); //Used  in constructor
      // }}}--------------------------------------------------------------------
      
      //Output Files
      File *etafile;
      File *phifile;
      File *phyphifile;
      File *horvelfile;
      File *parametersfile;
      
      //DirichletBC
      DirichletBC *dbc;
      bool dbcapply;


      // Assert and NotImplemented Functions
    
      template<class DI> inline void Assert(DI x,unsigned int line,
                                            std::string file )
      {
        if (!(x))                  
          {                                 
            std::ostringstream ss;                                          
            ss<<"ERROR!! Assert " << x << " failed\n";                     
            fputs(ss.str().c_str(),stderr);                                 
            ss<<" on line " << line  << "\n";                           
            fputs(ss.str().c_str(),stderr);                                 
            ss<<" in file " << file << "\n";                            
            fputs(ss.str().c_str(),stderr);                                 
            abort();                                                        
          }
      };
      
      inline void NotImplemented(std::string s,
                                 unsigned int line,
                                 std::string file)
      { 
        fputs(s.c_str(),stderr);
        fputs("\n Not implemented yet, failed\n",stderr); 
        Assert(0,line,file);                                     
      };
    
    }; 
  }
}
#endif 
