// {{{ Copyright (C) 2009 Nuno David Lopes.-----------------------------------
//
// e-mail: ndl@ptmat.fc.ul.pt
// http://ptmat.fc.ul.pt/~ndl/
//
// ---------------------------------------------------------------------------
// Licensed under the GNU LGPL Version 3.0 or later.
//
// ---------------------------------------------------------------------------
//  dolfwave.h
//
// First added:  2009-04-14
// Compiled with DOLFIN:0.9.2
// }}}------------------------------------------------------------------------

#include"DolfwaveClass.h"

using namespace dolfin;
using namespace dolfin::dolfwave;


// {{{ FunctionSpaceInit------------------------------------------------

void dolfin::dolfwave::Dolfwave::FunctionSpaceInit(dolfin::Mesh &mesh)
{
  if (FormName=="LPT2z0")//LPT2z
    V= new LPT2z0FunctionSpace(mesh);  
  else if (FormName=="Zhao")//Zhao
    V= new ZhaoFunctionSpace(mesh);
  else if (FormName=="LPTcdgI")
    V= new LPTcdgIFunctionSpace(mesh);
  else
    NotImplemented("FunctionSpaceInit form",__LINE__,__FILE__);

  V0= new SubSpace(*V,0);
  V1= new SubSpace(*V,1);
}

// }}}--------------------------------------------------------------------

// {{{ BilinearFormInit----------------------------------------------
void dolfin::dolfwave::Dolfwave::BilinearFormInit(dolfin::Function &h)
{
  
  
  if (FormName=="LPT2z0")//LPT2z0
    {
      a=new LPT2z0BilinearForm(*V,*V,h);//h=h0
    }
  else if (FormName=="Zhao")//Zhao
    {
      a=new ZhaoBilinearForm(*V,*V,h);
    }
  else
    NotImplemented("Bilinear Form Init",__LINE__,__FILE__);
}


void dolfin::dolfwave::Dolfwave::TestBilinearFormInit(dolfin::Function &h,dolfin::Function &n,
                                                         dolfin::Function &tau,dolfin::Function &e)
{
  if (FormName=="LPTcdgI")//Zhao
    {
      a=new LPTcdgIBilinearForm(*V,*V,h,n,tau,e);
    }
  else
    NotImplemented("Bilinear Form Init",__LINE__,__FILE__);
}
// }}}---------------------------------------------------------------

// {{{ LinearFormsInit ---------------------------------------------

void dolfin::dolfwave::Dolfwave::LinearFormsInit(dolfin::Function &depth,
                                                 dolfin::Function &g,
                                                 dolfin::Function &sponge,
                                                 dolfin::Function &src)
{
  L=new dolfin::Form * [5];
  if (FormName=="LPT2z0")//LPT2z0
    {
      for(dolfin::uint i=0;i<5;i++)
        L[i]=new LPT2z0LinearForm(*V,depth,*eta[i],
                                  *phi[i],g,sponge,src);
    }
  else
    NotImplemented("Linear Forms Init",__LINE__,__FILE__);
  
}


void dolfin::dolfwave::Dolfwave::LinearFormsInit(dolfin::Function &depth,
                                                 dolfin::Function &g,
                                                 dolfin::Function &sp_eta,
                                                 dolfin::Function &sp_lap_eta,
                                                 dolfin::Function &sp_phi,
                                                 dolfin::Function &sp_lap_phi,
                                                 dolfin::Function &src)
{
  L=new dolfin::Form * [5];
  if (FormName=="Zhao")//Zhao
    {
      for(dolfin::uint i=0;i<5;i++)
        L[i]=new ZhaoLinearForm(*V,depth,*eta[i],*phi[i],g,
                                sp_eta,sp_lap_eta,sp_phi,
                                sp_lap_phi,src);
    }
  else 
    NotImplemented("Linear Forms Init Several sponges.",__LINE__,__FILE__);
}


void dolfin::dolfwave::Dolfwave::TestLinearFormsInit(dolfin::Function &depth,
                                                     dolfin::Function &g,
                                                     dolfin::Function &n,
                                                     dolfin::Function &tau,
                                                     dolfin::Function &e)
{
  L=new dolfin::Form * [5];
  if (FormName=="LPTcdgI")
    {
      for(dolfin::uint i=0;i<5;i++)
        L[i]=new LPTcdgILinearForm(*V,depth,*eta[i],*phi[i],
                                   g,n,tau,e);
    }
  else
    NotImplemented("Linear Forms Init",__LINE__,__FILE__);
  
}
// }}} --------------------------------------------------------


// {{{ FunctionsInit ------------------------------------------------------- 
void dolfin::dolfwave::Dolfwave::FunctionsInit( )
{
  eta_phi=new dolfin::Function *  [5];
  eta=new dolfin::Function * [5];
  phi=new dolfin::Function * [5];
  for (dolfin::uint i=0;i<5;i++)
    {
      eta_phi[i]=new dolfin::Function(*V);
      eta[i]=new dolfin::Function(*V0);
      phi[i]=new dolfin::Function(*V1);
      eta_phi[i]->vector().zero();
      *eta[i]=(*eta_phi[i])[0];//updating subfunction information
      *phi[i]=(*eta_phi[i])[1];//updating subfunction information
    }
}
// }}}---------------------------------------------------------------------

// {{{ VectorInit ------------------------------------------------
void dolfin::dolfwave::Dolfwave::VectorInit( )
{
  pcrk_b=new Vector [6];//predictor-correct and runge-kutta auxiliar righ-hand vectors  
  for(dolfin::uint i=0;i<6;i++)
    pcrk_b[i]=eta_phi[0]->vector(); //pcrk_b[5] initialization 
  
}
// }}}----------------------------------------------------------

// {{{ InitialCondition P1XP1 :: From Files
void dolfin::dolfwave::Dolfwave::InitialCondition( std::string eta_file, std::string phi_file) 
{
  message("Reading values for initial conditions");
  message("Only for raw files\n");
  message("Working only with P1 elements");
  message("Functions SubSpaces must be the same.");

  dolfin::uint N=0;

  std::ifstream fich_eta(eta_file.c_str());
  Assert(fich_eta.is_open(),__LINE__,__FILE__);
  std::ifstream fich_phi(phi_file.c_str());
  Assert(fich_phi.is_open(),__LINE__,__FILE__);
  
  fich_eta>>N;
  Assert(N,__LINE__,__FILE__);
  fich_phi>>N;
  Assert(N,__LINE__,__FILE__);
  cout<<N<<endl;
 
  double etavalues[N];
  double phivalues[N];
 
  double aux=0.0;
  for(dolfin::uint i=0;i<N;i++)
    {
      fich_eta>>aux;
      etavalues[i]=aux;
      fich_phi>>aux;
      phivalues[i]=aux;
    }
  
  fich_eta.close();
  fich_phi.close();

  
  Vector vector_eta(N);
  Vector vector_phi(N);
 
  Function eta_init(*V0,vector_eta);
  Function phi_init(*V0,vector_phi);
  vector_eta.set(etavalues);
  vector_phi.set(phivalues);

  Matrix ML2P;
  Vector b_init;
  L2ProjBilinearForm bil2p(*V,*V);
  L2ProjLinearForm lil2p(*V);
  lil2p.eta_init=eta_init;lil2p.phi_init=phi_init;
  assemble(ML2P,bil2p);
  assemble(b_init,lil2p);
  LUSolver solver;
  solver.solve(ML2P,eta_phi[0]->vector(),b_init);
  
  //initialize all "step" functions with initial conditions
  for (dolfin::uint i=1;i<5;i++)
    eta_phi[i]->vector()=eta_phi[0]->vector();
  for(dolfin::uint i=0;i<5;i++)
    {
      (*eta[i])=(*eta_phi[0])[0];//updating subfunction information
      (*phi[i])=(*eta_phi[0])[1];//updating subfunction information
    }
  
  message("***********End Reading the Initial Conditions**********\n");
}
// }}}----------------------------------------------------------

// {{{ InitialCondition  Most Used
void dolfin::dolfwave::Dolfwave::InitialCondition(Function &eta_init, Function& phi_init)
{
  Matrix ML2P;
  Vector b_init;
  Form *bil2p;
  Form *lil2p;
  
  if (FormName=="LPTcdgI")
    {
      message("Test P2 INITIAL CONDITIONS FOR LPTcdgI");
      bil2p= new L2ProjP2BilinearForm(*V,*V);
      lil2p= new L2ProjP2LinearForm (*V,eta_init,phi_init);
    }
  else
    {
      bil2p= new L2ProjBilinearForm(*V,*V);
      lil2p= new L2ProjLinearForm (*V,eta_init,phi_init);
    }
  
  assemble(ML2P,*bil2p);
  assemble(b_init,*lil2p);
  LUSolver solver;
  solver.solve(ML2P,eta_phi[0]->vector(),b_init);
  //initialize all "step" functions with initial conditions
  for (dolfin::uint i=1;i<5;i++)
    eta_phi[i]->vector()=eta_phi[0]->vector();
  for(dolfin::uint i=0;i<5;i++)
    {
      (*eta[i])=(*eta_phi[0])[0];//updating subfunction information
      (*phi[i])=(*eta_phi[0])[1];//updating subfunction information
    }
  
}

// }}}--------------------------------------------------------


// {{{MatricesAssemble.. ---------------------
void dolfin::dolfwave::Dolfwave::MatricesAssemble( )
{
 
  // This is temporary
      M= new Matrix();
      Mf= new Matrix();
      message("Matrices should be DolfwaveClass Members also");
      assemble(*M,*a);
      // We should use a copy constructor
      // Matrix Mf(M)? How to use it?
      assemble(*Mf,*a);
      
      if(dbcapply) //apply dirichlet boundary condition
        dbc->apply(*Mf,pcrk_b[5]);       
}

// }}} ----------------------------------------------------

