// {{{ Copyright (C) 2009 Nuno David Lopes.-----------------------------------
//
// e-mail: ndl@ptmat.fc.ul.pt
// http://ptmat.fc.ul.pt/~ndl/
//
// ---------------------------------------------------------------------------
// Licensed under the GNU LGPL Version 3.0 or later.
//
// ---------------------------------------------------------------------------
//  dolfwave.h
//
// First added:  2009-04-14
// Compiled with DOLFIN:0.9.2
// }}}------------------------------------------------------------------------

#include"DolfwaveClass.h"
#include "forms/SymProj.h"
#include "forms/Proj.h"
#include "forms/GradP0.h"
#include "forms/ZhaoPotentialTrans.h"

#include <fstream>

using namespace dolfin;
using namespace dolfin::dolfwave;

// {{{ Plot Functions---------------------------------------

/* 
preview the Function func with viper
*/
void dolfin::dolfwave::Dolfwave::ViperPlot(const dolfin::Function &func) const
{
  std::stringstream command;
  FunctionPlotData w(func);
  message("Plotting %s (%s), press 'q' to continue...",
          w.name().c_str(), w.label().c_str());
  
  // Save to file
  const std::string filename = dolfin_get("plot file name");
  File file(filename, true);
  
  file << w;
  // Plot data from file
  command << "viper --mode=auto " << filename;//<<"&";
  if (system(command.str().c_str()) != 0)
    warning("Unable to plot.");
  
  
  // I Don't like the solution
  // Aks Viper authors to incorporate the feature
  /*
    if ((step%gap)==gap-1)
    {
    command << "pkill viper";
    if (system(command.str().c_str()) != 0)
    warning("Unable to kill viper.");
    }
  */
}

/* 
Preview the function func with Xd3d
*/
void dolfin::dolfwave::Dolfwave::Xd3dXyzPlot(const dolfin::Function &func) const
{
  std::stringstream command;
  message("Preview with  xd3d 'xyz' file, Press 'q' to resume");
  
  // Save to file
  File file("dolfin_xd3d.xyz", true);
  
  file << func;
  // Plot data from file
  command << "xd3d -bord=3 -iso=z -iso=1 dolfin_xd3d000000.xyz";
  if (system(command.str().c_str()) != 0)
    warning("Unable to plot.");
  
}

/*
 Plot eta_phi functions
 show=true will show a preview the surface elevation eta with viper
 surface=true will save eta in raw file format
 potential=true  will save (potential or pseudo potential) phi in raw format 
*/ 


void dolfin::dolfwave::Dolfwave::Plot(const bool show_eta,
                                      const bool show_phi,
                                      const bool surface,
                                      const bool potential) const
{
  if(show_eta)
    {
      if (Plotter=="viper")
        ViperPlot((*eta_phi[4])[0]);
      else if (Plotter=="xd3d_xyz")
        Xd3dXyzPlot((*eta_phi[4])[0]);
      else
        message("NOT IMPLEMENTED->RESUMING WITHOUT PREVIEW");
    }     
 if(show_phi)
   {
     if (Plotter=="viper")
       ViperPlot((*eta_phi[4])[1]);
     else if (Plotter=="xd3d_xyz")
       Xd3dXyzPlot((*eta_phi[4])[1]);
     else
       message("NOT IMPLEMENTED->RESUMING WITHOUT PREVIEW");
   }
 
 if(surface)
   *etafile<< (*eta_phi[4])[0];
 if(potential)
   *phifile<< (*eta_phi[4])[1];
}
/*
Saving and/or preview with viper(show=true) the depth function
It uses the symmetric projection of function
*/
void dolfin::dolfwave::Dolfwave::DepthPlot(const Mesh &mesh ,Function &depth,const bool show) const
{
  SymProjFunctionSpace W(mesh);
  SymProjBilinearForm a_sp(W,W);
  SymProjLinearForm Lsp(W);
  Lsp.f=depth;
  VariationalProblem symproj(a_sp, Lsp);
  Function d;
  symproj.solve(d);
  File file_raw("output/depth.raw");
  File file_vtk("output/depth.pvd");
  file_raw<<d;//just for post-processor
  file_vtk<<d;//just for post-processor 
  if (show) 
    ViperPlot(d);
}

/*
Saving and/or preview with viper(show=true) a user defined function
(Sponge is used in this case)
 */
void dolfin::dolfwave::Dolfwave::SpongePlot(const Mesh &mesh ,Function &sponge,const bool show) const
{
  ProjFunctionSpace W(mesh);
  ProjBilinearForm a(W,W);
  ProjLinearForm L(W);
  L.f=sponge;
  VariationalProblem proj(a, L);
  Function s;
  proj.solve(s);
  File file_raw("output/sponge.raw");
  File file_vtk("output/sponge.pvd");
  file_raw<<s;//just for post-processor
  file_vtk<<s;//just for post-processor
  if (show) 
    ViperPlot(s);
}

/*
Saving and/or preview Horizontal Gradient of phi
*/

void dolfin::dolfwave::Dolfwave::HorVelPlot(const Mesh &mesh ,Function**& phi ,const bool show, const bool save) const
{
  message("Horizontal Velocities calculation");
  GradP0FunctionSpace W(mesh);
  GradP0BilinearForm a(W,W);
  GradP0LinearForm L(W);
  L.phi=(*phi[4]);
  VariationalProblem vel(a, L);
  Function uv;
  vel.solve(uv);
  if(save)
    *horvelfile<<uv;//just for post-processor    
  if(show)
    {
      if (Plotter=="viper")
        ViperPlot(uv);
      else
        message("Not implemented yet");
          
    }

}

/*
Saving and/or preview Physical Potential
Implented only for Zhao Teng and Cheng Model
*/
void dolfin::dolfwave::Dolfwave::PhyPhiPlot(const Mesh &mesh ,Function &depth,Function**& phi ,Function &flux_switch,const bool show, const bool save) const
{
  if(FormName=="Zhao")
    {
      message("Physical Potential calculation");
      ZhaoPotentialTransFunctionSpace W(mesh);
      ZhaoPotentialTransBilinearForm a(W,W);
      ZhaoPotentialTransLinearForm L(W);
      L.Phi=*phi[4];
      L.h=depth;
      L.flux_switch=flux_switch;
      FacetNormal n;
      L.n=n;

      VariationalProblem vel(a, L);
      Function pp;
      vel.solve(pp);
      if(save)
        *phyphifile<<pp;//just for post-processor    
      if(show)
        {
          if (Plotter=="viper")
            ViperPlot(pp);
          else
            message("Not implemented yet");
        }
    }
  else 
    message("Not implemented yet, or USELESS");
}




//Save the function profile in a Specific Point
void dolfin::dolfwave::Dolfwave::PointPlot(const double x ,const double y, Function**& func,const std::string filename) const
{
  double value=0.0;
  std::ofstream fp(filename.c_str(),std::ios_base::app); //open and appending 
  double p[2]={x,y};
  func[4]->eval(&value,p);
  fp<< Time <<" " <<value<<std::endl;
}
// }}}----------------------------------------------------
