// {{{ Copyright (C) 2009 Nuno David Lopes.-----------------------------------
//
// e-mail: ndl@ptmat.fc.ul.pt
// http://ptmat.fc.ul.pt/~ndl/
//
// ---------------------------------------------------------------------------
// Licensed under the GNU LGPL Version 3.0 or later.
//
// ---------------------------------------------------------------------------
//  PredictorCorrector.cpp
//
// First added:  2009-04-09
// Compiled with DOLFIN:0.9.2
// }}}------------------------------------------------------------------------


#include <dolfin.h>
#include "dolfwave.h"

#include "PredictorCorrector.h"

using namespace dolfin;
using namespace dolfin::dolfwave;


// {{{  Predictor Corrector-Method---------------------------
PredCorr::PredCorr()
{
  // Constants for the 4th order Adams-Basfourth-Molton 
  p_cts[0]=55./24.;
  p_cts[1]=-59./24.;
  p_cts[2]=37./24.;
  p_cts[3]=-9./24.;
  
  c_cts[0]=9./24.;
  c_cts[1]=19./24.;
  c_cts[2]=-5./24.;
  c_cts[3]=1./24.;

  multicorrector=false;
  PcEpsilon=1.0;
  
  disp();
}

PredCorr::PredCorr(FunctionSpace &V0,Function**& eta_phi, double eps)
{
  // Constants for the 4th order Adams-Basfourth-Molton 
  p_cts[0]=55./24.;
  p_cts[1]=-59./24.;
  p_cts[2]=37./24.;
  p_cts[3]=-9./24.;
  
  c_cts[0]=9./24.;
  c_cts[1]=19./24.;
  c_cts[2]=-5./24.;
  c_cts[3]=1./24.;
  
  multicorrector=true;
  l2error=1.0;
  PcEpsilon=eps;

  eta_old= new Function(V0); *eta_old=(*eta_phi[3])[0];
  phi_old= new Function(V0); *phi_old=(*eta_phi[3])[1];
  eta_new= new Function(V0); 
  phi_new= new Function(V0); 
  ML2N= new L2NormFunctional(*eta_new,*eta_old,*phi_new,*phi_old);
  disp();
};

void PredCorr::pred_assemble(Vector*& pcrk_b,  
                             const dolfin::Form &L0,  
                             const dolfin::Form &L1,  
                             const dolfin::Form &L2,
                             const dolfin::Form &L3,
                             Matrix &M,
                             GenericVector& y,
                             double timestep)
{
  assemble(pcrk_b[0],L0);//f(y[0])
  assemble(pcrk_b[1],L1);//f(y[1])
  assemble(pcrk_b[2],L2);//f(y[2])
  assemble(pcrk_b[3],L3);//f(y[3])
  M.mult(y,pcrk_b[5]);//pcrk_b[5]=M*y[3]
  for(dolfin::uint j=0;j<4;j++)
    pcrk_b[5].axpy((timestep*p_cts[j]),pcrk_b[j]);
  
};

void PredCorr::corr_assemble(Vector*& pcrk_b,  
                             const dolfin::Form &L1,  
                             const dolfin::Form &L2,  
                             const dolfin::Form &L3,
                             const dolfin::Form &L4,
                             Matrix &M,
                             GenericVector& y,
                             double timestep)
{
    assemble(pcrk_b[1],L1);//f(y[1])
    assemble(pcrk_b[2],L2);//f(y[2])
    assemble(pcrk_b[3],L3);//f(y[3])
    assemble(pcrk_b[4],L4);//f(y[4]) //using y[4]^0 from the predictor step
    M.mult(y,pcrk_b[5]);//pcrk_b[5]=M*y[3]
    for(dolfin::uint j=1;j<5;j++)
      pcrk_b[5].axpy((timestep*c_cts[j-1]),pcrk_b[j]);
  };
  
void PredCorr::disp()
  {
    cout<<"**************************************"<<endl;
    cout<<"ABM Predict.Correct. 4 order constants"<<endl;
    cout<<"******************PRED****************"<<endl;
    for(dolfin::uint i=0;i<4;i++)
      cout<<p_cts[i]<<"\t";
    cout<<endl;
    cout<<"******************CORR****************"<<endl;
    for(dolfin::uint i=0;i<4;i++)
      cout<<c_cts[i]<<"\t";
    cout<<endl;
    if (multicorrector)
      cout<<"Multicorrector Steps=True"<<endl;
    else
      cout<<"Multicorrector Steps=False"<<endl;
    cout<<"**************************************"<<endl;
  }; 

void PredCorr::update(Function &eta0,Function &eta1,Function &eta2,Function &eta3,Function &eta4,
                      Function &phi0,Function &phi1,Function &phi2,Function &phi3,Function &phi4,
                      Function &eta_phi0,Function &eta_phi1,Function &eta_phi2,Function &eta_phi3,Function &eta_phi4)
{
  
  eta_phi0.vector()=eta_phi1.vector();
  eta_phi1.vector()=eta_phi2.vector();
  eta_phi2.vector()=eta_phi3.vector();
  eta_phi3.vector()=eta_phi4.vector();
     
  
  eta0=eta_phi0[0];phi0=eta_phi0[1];//updating function information
  eta1=eta_phi1[0];phi1=eta_phi1[1];//updating function information
  eta2=eta_phi2[0];phi2=eta_phi2[1];//updating function information
  eta3=eta_phi3[0];phi3=eta_phi3[1];//updating function information
  eta4=eta_phi4[0];phi4=eta_phi4[1];//updating function information
}

void PredCorr::update(Function**& eta,Function**& phi,Function**& eta_phi)
{
  for(dolfin::uint i=0;i<4;i++)
    eta_phi[i]->vector()=eta_phi[i+1]->vector();
  for(dolfin::uint i=0;i<5;i++)
    {
      *eta[i]=(*eta_phi[i])[0];
      *phi[i]=(*eta_phi[i])[1];//updating function information
    }
}

void PredCorr::sbfctn_update(Function &eta,Function &phi,Function &eta_phi)
{
  
  
  eta=eta_phi[0];phi=eta_phi[1];//updating function information
}

// {{{PREDICTOR STEP SOLVER------------------------------------------
/*
with and without Dirichlet boundary conditions
*/


void PredCorr::pred_solve(Vector*& pcrk_b,
                const dolfin::Form &L0,
                const dolfin::Form &L1,
                const dolfin::Form &L2,
                const dolfin::Form &L3,
                Matrix &M,
                Function &eta_phi3,
                Function &eta_phi4,
                Function &eta4,
                Function &phi4,
                double  timestep,
                DirichletBC &dbc,
                LUSolver &lu,
                Matrix &Mf,
                double &t)
{
  pred_assemble(pcrk_b,L0,L1,L2,L3,M,
                eta_phi3.vector(),timestep);
  dbc.apply(Mf,pcrk_b[5]);
  lu.factorized_solve(eta_phi4.vector(),pcrk_b[5]);//eta_phi4^0 is calculated
  sbfctn_update(eta4,phi4,eta_phi4);//updating subfunction information
  
};

void PredCorr::pred_solve(Vector*& pcrk_b,
                          const dolfin::Form &L0,
                          const dolfin::Form &L1,
                          const dolfin::Form &L2,
                          const dolfin::Form &L3,
                          Matrix &M,
                          Function &eta_phi3,
                          Function &eta_phi4,
                          Function &eta4,
                          Function &phi4,
                          double  timestep,
                          LUSolver &lu,
                          double &t)
{
  pred_assemble(pcrk_b,L0,L1,L2,L3,M,
                eta_phi3.vector(),timestep);
  lu.factorized_solve(eta_phi4.vector(),pcrk_b[5]);//eta_phi4^0 is calculated
  sbfctn_update(eta4,phi4,eta_phi4);//updating subfunction information
  
};
// }}}--------------------------------------------------------------------------- 

// {{{CORRECTOR STEP SOLVER------------------------------------------
/*
with and without Dirichlet boundary conditions
*/

void PredCorr::corr_solve(Vector *&pcrk_b,
                const dolfin::Form &L1,
                const dolfin::Form &L2,
                const dolfin::Form &L3,
                const dolfin::Form &L4,
                Matrix &M,
                Function &eta_phi3,
                Function &eta_phi4,
                Function &eta4,
                Function &phi4,
                double  timestep,
                DirichletBC &dbc,
                LUSolver &lu,
                Matrix &Mf,
                double &t)
{
  corr_assemble(pcrk_b,L1,L2,L3,L4,M,
                eta_phi3.vector(),timestep);
  dbc.apply(Mf,pcrk_b[5]);
  lu.factorized_solve(eta_phi4.vector(),pcrk_b[5]);
  sbfctn_update(eta4,phi4,eta_phi4);//updating subfunction information
};

void PredCorr::corr_solve(Vector *&pcrk_b,
                          const dolfin::Form &L1,
                          const dolfin::Form &L2,
                          const dolfin::Form &L3,
                          const dolfin::Form &L4,
                          Matrix &M,
                          Function &eta_phi3,
                          Function &eta_phi4,
                          Function &eta4,
                          Function &phi4,
                          double  timestep,
                          LUSolver &lu,
                          double &t)
{
  corr_assemble(pcrk_b,L1,L2,L3,L4,M,
                eta_phi3.vector(),timestep);
  lu.factorized_solve(eta_phi4.vector(),pcrk_b[5]);
  sbfctn_update(eta4,phi4,eta_phi4);//updating subfunction information
};
// }}}--------------------------------------------------------------------------- 

// {{{ GENERIC PREDICTOR CORRECTOR SOLVER------------------------------------------
/*
with and without Dirichlet boundary conditions
*/

void PredCorr::solve(Vector*& pcrk_b,
                     dolfin::Form**& L,
                     Matrix &M,
                     Function**& eta_phi,
                     Function**& eta,
                     Function**& phi,
                     double  timestep,
                     DirichletBC &dbc,
                     LUSolver &lu,
                     Matrix &Mf,
                     double &t)
{
  pred_solve(pcrk_b,*L[0],*L[1],*L[2],*L[3],
             M,*eta_phi[3],*eta_phi[4],*eta[4],*phi[4],
             timestep,dbc,lu,Mf,t);
  if(!multicorrector)
    corr_solve(pcrk_b,*L[1],*L[2],*L[3],*L[4],
               M,*eta_phi[3],*eta_phi[4],*eta[4],*phi[4],
               timestep,dbc,lu,Mf,t);
  else
    {
      while (l2error>PcEpsilon)
        {
          *eta_old=(*eta_phi[4])[0];*phi_old=(*eta_phi[4])[1];
          
          corr_solve(pcrk_b,*L[1],*L[2],*L[3],*L[4],
                     M,*eta_phi[3],*eta_phi[4],*eta[4],*phi[4],
                     timestep,dbc,lu,Mf,t);
          
          *eta_new=(*eta_phi[4])[0];*phi_new=(*eta_phi[4])[1];
          l2error=assemble(*ML2N); 
          l2error=sqrt(l2error);
          if (isnan(l2error)||isinf(l2error))
            {
              fputs ("NAN or INF values of multi-step corrector \n",stderr);
              abort();
            }
          
          message("*************************Corrector L2 error=%e",l2error);
        }
      l2error=1.0;//reset l2error for the next step;
    };
  
  update(eta,phi,eta_phi);
}

void PredCorr::solve(Vector*& pcrk_b,
                     dolfin::Form**& L,
                     Matrix &M,
                     Function**& eta_phi,
                     Function**& eta,
                     Function**& phi,
                     double  timestep,
                     LUSolver &lu,
                     double &t)
{
  pred_solve(pcrk_b,*L[0],*L[1],*L[2],*L[3],
             M,*eta_phi[3],*eta_phi[4],*eta[4],*phi[4],
             timestep,lu,t);
  if(!multicorrector)
    corr_solve(pcrk_b,*L[1],*L[2],*L[3],*L[4],
               M,*eta_phi[3],*eta_phi[4],*eta[4],*phi[4],
               timestep,lu,t);
  else
    {
      while (l2error>PcEpsilon)
        {
          *eta_old=(*eta_phi[4])[0];*phi_old=(*eta_phi[4])[1];
          
          corr_solve(pcrk_b,*L[1],*L[2],*L[3],*L[4],
                     M,*eta_phi[3],*eta_phi[4],*eta[4],*phi[4],
                     timestep,lu,t);
          
          *eta_new=(*eta_phi[4])[0];*phi_new=(*eta_phi[4])[1];
          l2error=assemble(*ML2N); 
          l2error=sqrt(l2error);
          if (isnan(l2error)||isinf(l2error))
            {
              fputs ("NAN or INF values of multi-step corrector \n",stderr);
              abort();
            }


          message("*************************Corrector L2 error=%e",l2error);
        }
      l2error=1.0;//reset l2error for the next step;
        };
  
  update(eta,phi,eta_phi);
}


// }}}--------------------------------------------------------------------------- 

// }}}--------------------------------------------------------------------------- 

